import logging
import re

import lxml.etree as ET

from excel2moodle.core.question import Parametrics

loggerObj = logging.getLogger(__name__)


class MediaCall:
    def __init__(self, callString: str, divId: str) -> None:
        self.call: str = re.sub(r"\(", f'("{divId}", ', callString)
        self._scriptElem = ET.Element("script")
        self._scriptElem.text = callString
        self.divId = divId

    @property
    def element(self) -> ET.Element:
        return self._scriptElem

    def update(self, parametrics: Parametrics, variant: int = 0) -> None:
        updatedCall = self.call
        for var, val in parametrics.variables.items():
            updatedCall = re.sub(rf",\s+{var}\s*", f", {val[variant]}", updatedCall)
        self._scriptElem.text = updatedCall
        loggerObj.info("Inserted function call: %s", updatedCall)



"""This *extra* subpackage conains standalone scripts

The modules inside *extra* can be run standalone, but are planned to be available from the main Window as well

To run a script execute the following: ``python -m excel2moodle.extra.SCRIPT``
Note that there is no ``.py`` at the end!!
"""



"""This module provides functions to query the GitLab API for project information."""

import json
import sys
import urllib.request


def get_latest_tag(project_id: str) -> str | None:
    """Queries the GitLab API for the latest tag of a project.

    Args:
        project_id: The URL-encoded project ID (e.g., 'jbosse3%2Fexcel2moodle').

    Returns:
        The name of the latest tag.

    """
    url = f"https://gitlab.com/api/v4/projects/{project_id}/repository/tags"
    try:
        with urllib.request.urlopen(url) as response:
            if response.status == 200:
                data = json.loads(response.read().decode())
                if data:
                    return data[0]["name"]
    except urllib.error.URLError as e:
        print(f"Error fetching latest tag: {e}", file=sys.stderr)
    return None


def get_changelog(project_id: str, branch: str = "master") -> str:
    """Queries the GitLab API for the content of the CHANGELOG.md file.

    Args:
        project_id: The URL-encoded project ID (e.g., 'jbosse3%2Fexcel2moodle').
        branch: The branch to get the file from.

    Returns:
        The content of the CHANGELOG.md file.

    """
    url = f"https://gitlab.com/api/v4/projects/{project_id}/repository/files/CHANGELOG.md/raw?ref={branch}"
    try:
        with urllib.request.urlopen(url) as response:
            if response.status == 200:
                return response.read().decode()
    except urllib.error.URLError as e:
        print(f"Error fetching changelog: {e}", file=sys.stderr)
    return ""



"""Numerical question multi implementation."""

from types import UnionType
from typing import ClassVar

import lxml.etree as ET

from excel2moodle.core.globals import (
    Tags,
    TextElements,
    XMLTags,
)
from excel2moodle.core.parser import QuestionParser
from excel2moodle.core.question import ParametricQuestion, Parametrics


class NFMQuestion(ParametricQuestion):
    mandatoryTags: ClassVar[dict[Tags, type | UnionType]] = {
        Tags.RESULT: str,
        Tags.BPOINTS: str,
    }

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.answerElement: ET.Element
        self.answerElementWrongSign: ET.Element

    def getUpdatedElement(self, variant: int = 1) -> ET.Element:
        """Update and get the Question Elements to reflect the version.

        `NFMQuestion` updates the answer Elements.
        `ParametricQuestion` updates the bullet points.
        `Question` returns the Element.
        """
        result = self.parametrics.getResult(variant)
        tolerance = round(self.rawData.get(Tags.TOLERANCE) * result, 3)
        self.answerElement.find(XMLTags.TEXT).text = str(result)
        self.answerElementWrongSign.find(XMLTags.TEXT).text = str(result * (-1))
        self.answerElement.find(XMLTags.TOLERANCE).text = str(tolerance)
        self.answerElementWrongSign.find(XMLTags.TOLERANCE).text = str(tolerance)
        return super().getUpdatedElement(variant)


class NFMQuestionParser(QuestionParser):
    def __init__(self) -> None:
        super().__init__()
        self.feedBackList = {XMLTags.GENFEEDB: Tags.GENERALFB}
        self.question: NFMQuestion

    def setup(self, question: NFMQuestion) -> None:
        self.question: NFMQuestion = question
        super().setup(question)
        module = self.settings.get(Tags.IMPORTMODULE)
        if module and Parametrics.astEval.symtable.get(module, None) is None:
            Parametrics.astEval(f"import {module}")
            imported = Parametrics.astEval.symtable.get(module)
            self.logger.warning("Imported '%s' to Asteval symtable.", module)

    def _parseAnswers(self) -> list[ET.Element]:
        variables = self.question.bulletList.getVariablesDict(self.question)
        self.question.parametrics = Parametrics(
            self.rawData.get(Tags.EQUATION),
            self.rawData.get(Tags.FIRSTRESULT),
            self.question.id,
        )
        self.question.parametrics.variables = variables
        self.question.answerElement = self.getNumericAnsElement()
        self.question.answerElementWrongSign = self.getNumericAnsElement(
            fraction=self.rawData.get(Tags.WRONGSIGNPERCENT),
            feedback=self.rawData.get(Tags.WRONGSIGNFB),
            feedbackStyle=TextElements.SPANORANGE,
        )
        return [self.question.answerElement, self.question.answerElementWrongSign]

    def _finalizeParsing(self) -> None:
        return super()._finalizeParsing()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'variantDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.8.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QDialog,
    QDialogButtonBox, QFormLayout, QFrame, QLabel,
    QSizePolicy, QSpinBox, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setRowWrapPolicy(QFormLayout.RowWrapPolicy.WrapLongRows)
        self.formLayout.setLabelAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.formLayout.setFormAlignment(Qt.AlignmentFlag.AlignCenter)
        self.formLayout.setHorizontalSpacing(20)
        self.formLayout.setVerticalSpacing(10)
        self.formLayout.setContentsMargins(-1, -1, 10, -1)
        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)

        self.spinBox = QSpinBox(Dialog)
        self.spinBox.setObjectName(u"spinBox")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.spinBox)

        self.label_2 = QLabel(Dialog)
        self.label_2.setObjectName(u"label_2")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_2)

        self.checkBox = QCheckBox(Dialog)
        self.checkBox.setObjectName(u"checkBox")
        self.checkBox.setChecked(True)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.checkBox)

        self.line = QFrame(Dialog)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.line)

        self.line_2 = QFrame(Dialog)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.HLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.line_2)

        self.catText = QLabel(Dialog)
        self.catText.setObjectName(u"catText")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.catText)

        self.catLabel = QLabel(Dialog)
        self.catLabel.setObjectName(u"catLabel")

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.catLabel)

        self.qtext = QLabel(Dialog)
        self.qtext.setObjectName(u"qtext")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.qtext)

        self.qLabel = QLabel(Dialog)
        self.qLabel.setObjectName(u"qLabel")

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.qLabel)

        self.label_3 = QLabel(Dialog)
        self.label_3.setObjectName(u"label_3")

        self.formLayout.setWidget(5, QFormLayout.LabelRole, self.label_3)

        self.idLabel = QLabel(Dialog)
        self.idLabel.setObjectName(u"idLabel")

        self.formLayout.setWidget(5, QFormLayout.FieldRole, self.idLabel)


        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Select question variant", None))
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Remember for all of this category", None))
        self.checkBox.setText("")
        self.catText.setText(QCoreApplication.translate("Dialog", u"Category:", None))
        self.catLabel.setText(QCoreApplication.translate("Dialog", u"TextLabel", None))
        self.qtext.setText(QCoreApplication.translate("Dialog", u"Question:", None))
        self.qLabel.setText(QCoreApplication.translate("Dialog", u"TextLabel", None))
        self.label_3.setText(QCoreApplication.translate("Dialog", u"ID:", None))
        self.idLabel.setText(QCoreApplication.translate("Dialog", u"TextLabel", None))
    # retranslateUi




"""This Python program helps to create Moodle questions in less time.

The aim is to put alle the information for the questions into a spreadsheet
file, and then parse it, to generate Moodle compliant XML-Files.
Furthermore this program lets you create a single ``.xml``-File with a selection
of questions, that then can be imported to a Moodle-Test.

Concept
=========
The concept is, to store the different questions into categories of similar
types and difficulties of questions, for each of which, a separated sheet in
the Spreadsheet document should be created.

There Should be a sheet called *"Kategorien"*, where an overview over the
different categories is stored.
This sheet stores The names and descriptions, for all categories.
The name have to be the same as the actual sheet names with the questions.
Furthermore the points used for grading, are set in the "Kategorien" sheet

Functionality
===============
* Parse multiple Choice Questions, each into one XML file
* Parse Numeric Questions, each into one XML file
* create single XML File from a selection of questions
"""

import logging
from importlib import metadata
from importlib.metadata import version

try:
    __version__ = version("excel2moodle")
except Exception:
    __version__ = "unknown"

e2mMetadata: dict = {}
if __package__ is not None:
    meta = metadata.metadata(__package__)
    e2mMetadata = {
        "version": __version__,
        "name": meta["name"],
        "description": meta["summary"],
        "author": meta["author"],
        "license": meta["license-expression"],
        "documentation": "https://jbosse3.gitlab.io/excel2moodle",
        "homepage": meta["project-url"].split()[1],
        "issues": "https://gitlab.com/jbosse3/excel2moodle/issues",
        "funding": "https://ko-fi.com/jbosse3",
        "API_id": "jbosse3%2Fexcel2moodle",
    }


mainLogger = logging.getLogger(__name__)



import logging
import re
from typing import Literal

import lxml.etree as ET

from excel2moodle.core import stringHelpers
from excel2moodle.core.exceptions import InvalidFieldException
from excel2moodle.core.globals import Tags, TextElements
from excel2moodle.core.question import ParametricQuestion, Parametrics
from excel2moodle.logger import LogAdapterQuestionID

loggerObj = logging.getLogger(__name__)


class BulletList:
    def __init__(
        self,
        rawBullets: list[str],
        qID: str,
        style: Literal["parameters", "raw"] = "parameters",
        template: str = "",
    ) -> None:
        self.rawBullets: list[str] = rawBullets
        self.element: ET.Element = ET.Element("ul")
        self.bullets: dict[str | int, BulletP | RawBulletP] = {}
        self.id = qID
        self.logger = LogAdapterQuestionID(loggerObj, {"qID": self.id})
        if style == "parameters":
            self._setupBullets(rawBullets, template=template)
        elif style == "raw":
            self._setupRawBullets(rawBullets)

    def appendBullets(
        self,
        bulletList: list[str],
        style: Literal["parameters", "raw"],
        template: str = "",
    ) -> None:
        self._setupRawBullets(bulletList)

    def update(self, parametrics: Parametrics, variant: int = 1) -> None:
        variables: dict[str, list[float]] = parametrics.variables
        for var, bullet in self.bullets.items():
            if isinstance(bullet, BulletP):
                bullet.update(value=variables[var][variant - 1])

    def getVariablesDict(self, question: ParametricQuestion) -> dict[str, list[float]]:
        """Read variabel values for vars in `question.rawData`.

        Returns
        -------
        A dictionary containing a list of values for each variable name

        """
        keyList = self.varNames
        dic: dict = {}
        for k in keyList:
            val = question.rawData[k.lower()]
            if isinstance(val, str):
                li = stringHelpers.getListFromStr(val)
                variables: list[float] = [float(i.replace(",", ".")) for i in li]
                dic[str(k)] = variables
            else:
                dic[str(k)] = [str(val)]
        loggerObj.debug("The following variables were provided: %s", dic)
        return dic

    @property
    def varNames(self) -> list[str]:
        names = [
            i for i in self.bullets if isinstance(i, str) and not i.startswith("raw")
        ]
        if len(names) > 0:
            self.logger.debug("returning Var names: %s", names)
            return names
        msg = "Bullet variable names not given."
        raise ValueError(msg)

    def _setupRawBullets(self, bps: list[str]) -> ET.Element:
        for i, bullet in enumerate(bps):
            self.bullets[f"raw{i}"] = RawBulletP(bulletStr=bullet)
            self.element.append(self.bullets[f"raw{i}"].element)
        return self.element

    def _setupBullets(self, bps: list[str], template: str = "") -> ET.Element:
        self.logger.debug("Formatting the bulletpoint list")
        varFinder = re.compile(r"\{(\w+)\}")
        bulletFinder = re.compile(
            r"^\s?(?P<desc>.*?)"
            r"(?:\s+(?P<var>[\S]+)\s*=\s*)"
            r"(?:(?P<val>[\S]+))"
        )
        unitFinder = re.compile(r".*?(?:=\s*[\S]+\s+(?P<unit>[\S]+)\s*$)")
        for i, item in enumerate(bps):
            match = re.search(bulletFinder, item)
            if match is None:
                self.logger.error("Couldn't find any bullets")
                msg = f"Couldn't decode the bullet point: {item}"
                raise InvalidFieldException(msg, self.id, Tags.BPOINTS)
            name = match.group("desc")
            var = match.group("var")
            value = match.group("val")
            unit = (
                match.group("unit")
                if (match := re.search(unitFinder, item)) is not None
                else ""
            )
            self.logger.info(
                "Decoded bulletPoint: name: %s, var: %s, - value: %s, - unit: %s.",
                name,
                var,
                value,
                unit,
            )
            if (match := re.search(varFinder, value)) is None:
                self.logger.debug("Got a normal bulletItem")
                num: float = float(value.replace(",", "."))
                bulletName = i + 1
            else:
                bulletName = match.group(1)
                num: float = 0.0
                self.logger.debug("Got an variable bulletItem, match: %s", match)
            # for userfriendliness because % would be the comment in latex
            if unit == "%":
                unit = r"\%"
            self.bullets[bulletName] = BulletP(
                name=name, var=var, unit=unit, value=num, template=template
            )
            self.element.append(self.bullets[bulletName].element)
        return self.element


class RawBulletP:
    def __init__(self, bulletStr: str) -> None:
        self.name = bulletStr
        self.element: ET.Element = TextElements.LISTITEM.create()
        self.element.text = bulletStr


class BulletP:
    def __init__(
        self,
        name: str,
        var: str,
        unit: str,
        value: float = 0.0,
        template: str = "",
    ) -> None:
        self.name: str = (
            name.replace(" ", "~")
            if template.startswith("\\(") and len(name.split()) > 1
            else name
        )
        self.var: str = var
        self.unit: str = unit
        self.element: ET.Element
        self.value: float = value
        mapper: dict[str, str] = {
            "<name>": self.name,
            "<var>": self.var,
            "<unit>": self.unit,
        }
        tmp = template
        for k, val in mapper.items():
            tmp = tmp.replace(k, str(val))
        self.bulletStr = tmp
        self.update(value=value)

    def update(self, value: float = 1) -> None:
        if not hasattr(self, "element"):
            self.element = TextElements.LISTITEM.create()
        valuestr = str(value).replace(".", r",\!")
        self.element.text = self.bulletStr.replace("<value>", valuestr)



"""Main Function to make the Package executable."""

import logging.config as logConfig
import signal
import sys
from pathlib import Path

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication

import excel2moodle
from excel2moodle import __version__, e2mMetadata, mainLogger
from excel2moodle.core import dataStructure
from excel2moodle.core.settings import Settings, Tags
from excel2moodle.extra import updateQuery
from excel2moodle.logger import loggerConfig
from excel2moodle.ui import appUi as ui


def main() -> None:
    app = QApplication(sys.argv)
    excel2moodle.isMain = True
    settings = Settings()
    logfile = Path(settings.get(Tags.LOGFILE)).resolve()
    e2mMetadata["logfile"] = logfile
    if logfile.exists() and logfile.is_file():
        logfile.replace(f"{logfile}.old")
    logConfig.dictConfig(config=loggerConfig)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    database: dataStructure.QuestionDB = dataStructure.QuestionDB(settings)
    window = ui.MainWindow(settings, database)
    database.window = window
    window.show()
    # Update check
    latestTag = updateQuery.get_latest_tag(e2mMetadata.get("API_id"))
    if latestTag is None:
        mainLogger.warning(
            "Couldn't check for Updates, maybe there is  no internet connection"
        )
    if latestTag is not None and latestTag.strip("v") != __version__:
        mainLogger.warning(
            "A new Update is available: %s --> %s ", __version__, latestTag
        )
        changelog = updateQuery.get_changelog(e2mMetadata.get("API_id"))
        # Delay showing the update dialog slightly
        QTimer.singleShot(100, lambda: window.showUpdateDialog(changelog, latestTag))

    for k, v in e2mMetadata.items():
        msg = f"{k:^14s}:  {v}"
        mainLogger.info(msg)
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



"""This Module holds small Helperfunctions related to string manipulation."""

from pathlib import Path

import lxml.etree as ET
from pandas import pandas


def getListFromStr(stringList: str | list[str | float]) -> list[str]:
    """Get a python List of strings from a semi-colon separated string."""
    stripped: list[str] = []
    li = []
    if isinstance(stringList, list):
        li = stringList
    elif isinstance(stringList, str):
        li = stringList.split(";")
    elif isinstance(stringList, float | int):
        li = [stringList]
    for i in li:
        s = i.strip() if isinstance(i, str) else str(i) if not pandas.isna(i) else None
        if s:
            stripped.append(s)
    return stripped


def getUnitsElementAsString(unit) -> None:
    def __getUnitEle__(name, multipl):
        unit = ET.Element("unit")
        ET.SubElement(unit, "multiplier").text = multipl
        ET.SubElement(unit, "unit_name").text = name
        return unit

    ET.Element("units")


def printDom(xmlElement: ET.Element, file: Path | None = None) -> None:
    """Prints the document tree of ``xmlTree`` to ``file``, if specified, else dumps to stdout."""
    documentTree = ET.ElementTree(xmlElement)
    if file is not None:
        if file.parent.exists():
            documentTree.write(
                file,
                xml_declaration=True,
                encoding="utf-8",
                pretty_print=True,
            )
    else:
        print(xmlElement.tostring())  # noqa: T201


def texWrapper(text: str | list[str], style: str) -> list[str]:
    r"""Put the strings inside ``text`` into a LaTex environment.

    if ``style == unit``: inside ``\\mathrm{}``
    if ``style == math``: inside ``\\( \\)``
    """
    answers: list[str] = []
    begin = ""
    end = ""
    if style == "math":
        begin = "\\("
        end = "\\)"
    elif style == "unit":
        begin = "\\(\\mathrm{"
        end = "}\\)"
    if isinstance(text, str):
        li = [begin]
        li.append(text)
        li.append(end)
        answers.append("".join(li))
    elif isinstance(text, list):
        for i in text:
            li = [begin]
            li.append(i)
            li.append(end)
            answers.append("".join(li))
    return answers



from pathlib import Path

import lxml.etree as ET

from excel2moodle.core.dataStructure import QuestionDB
from excel2moodle.core.settings import Settings, Tags

settings = Settings()

database = QuestionDB(settings)
excelFile = Path("test/TestQuestion.ods")
database.spreadsheet = excelFile
database.readCategoriesMetadata(excelFile)
database.initAllCategories(excelFile)
mcCategory = database.categories["MC1"]
nfmCategory = database.categories["NFM2"]
nfCategory = database.categories["NF3"]
database.parseCategoryQuestions(mcCategory)
database.parseCategoryQuestions(nfmCategory)
database.parseCategoryQuestions(nfCategory)


def test_unsetAnswerFeedback() -> None:
    q = mcCategory.questions[2]
    q.getUpdatedElement()
    true_feedback = ET.fromstring(
        q._element.find("answer").find("feedback").find("text").text
    )
    for ans in q._element.findall("answer"):
        if float(ans.get("fraction")) < 0.0:
            false_feedback = ET.fromstring(ans.find("feedback").find("text").text)
            break
    assert false_feedback.find("span").text == settings.get(Tags.FALSEANSFB)
    assert (
        false_feedback.find("span").text == "Your answer is sadly wrong, try again!!!"
    )
    assert true_feedback.find("span").text == settings.get(Tags.TRUEANSFB)
    assert true_feedback.find("span").text == "true"


def test_partialSetAnswerFB() -> None:
    q = mcCategory.questions[1]
    q.getUpdatedElement()
    for ans in q._element.findall("answer"):
        if float(ans.get("fraction")) > 0.0:
            trueAns = ans
        elif float(ans.get("fraction")) < 0.0:
            falseAns = ans
    trueFB = ET.fromstring(trueAns.find("feedback").find("text").text)
    falseFB = ET.fromstring(falseAns.find("feedback").find("text").text)
    assert trueFB.find("span").text == settings.get(Tags.TRUEANSFB)
    assert trueFB.find("span").text == "true"
    assert falseFB.find("span").text == settings.get(Tags.FALSEANSFB)
    assert falseFB.find("span").text == "Your answer is sadly wrong, try again!!!"


def test_nfAnswerFeedbacks() -> None:
    for question in nfCategory.questions.values():
        question.getUpdatedElement()
        tree = ET.Element("quiz")
        tree.append(question._element)
        answers = tree.find("question").findall("answer")
        wrongSignPerc = question.rawData.get(Tags.WRONGSIGNPERCENT)
        for ans in answers:
            trueAns = ans if ans.get("fraction") == "100" else None
        truefeedback = ET.fromstring(trueAns.find("feedback").find("text").text)
        print(truefeedback.find("span").text)
        assert truefeedback.find("span").text == settings.get(Tags.TRUEANSFB)


def test_nfmAnswerFeedbacks() -> None:
    for question in nfmCategory.questions.values():
        question.getUpdatedElement()
        tree = ET.Element("quiz")
        tree.append(question._element)
        answers = tree.find("question").findall("answer")
        wrongSignPerc = question.rawData.get(Tags.WRONGSIGNPERCENT)
        for ans in answers:
            if ans.get("fraction") == "100":
                trueAns = ans
            elif ans.get("fraction") == str(wrongSignPerc):
                wrongSignAns = ans
        truefeedback = ET.fromstring(trueAns.find("feedback").find("text").text)
        wrongSignEle = ET.fromstring(wrongSignAns.find("feedback").find("text").text)
        print(truefeedback.find("span").text)
        assert wrongSignEle.find("span").text == question.rawData.get(Tags.WRONGSIGNFB)
        assert truefeedback.find("span").text == settings.get(Tags.TRUEANSFB)



import logging
import re
from typing import TYPE_CHECKING

import lxml.etree as ET
import pandas as pd

if TYPE_CHECKING:
    from excel2moodle.core.question import Question
from excel2moodle.core.settings import Tags

loggerObj = logging.getLogger(__name__)


class Category:
    """Category stores a list of question. And holds shared information for all."""

    def __init__(
        self,
        name: str,
        description: str,
        dataframe: pd.DataFrame,
        settings: dict[str, float | str],
    ) -> None:
        """Instantiate a new Category object."""
        self.NAME = name
        match = re.search(r"\d+", str(self.NAME))
        n = int(match.group(0)) if match else 99
        self.n: int = n if n <= 99 and n >= 0 else 99
        self.desc = str(description)
        self.dataframe: pd.DataFrame = dataframe
        self.settings: dict[str, float | str] = settings if settings else {}
        self._questions: dict[int, Question]
        self.maxVariants: int | None = None
        loggerObj.info("initializing Category %s", self.NAME)

    @property
    def name(self) -> str:
        return self.NAME

    @property
    def points(self) -> float:
        return self.settings.get(Tags.POINTS)

    @points.setter
    def points(self, points: float) -> None:
        self.settings[Tags.POINTS] = points

    @property
    def id(self) -> str:
        return f"{self.n:02d}"

    @property
    def questions(self) -> dict:
        if not hasattr(self, "_questions"):
            msg = f"Category {self.id} doesn't contain any valid questions."
            raise ValueError(msg)
        return self._questions

    def appendQuestion(self, questionNumber: int, question) -> None:
        if not hasattr(self, "_questions"):
            self._questions: dict[int, Question] = {}
        self._questions[questionNumber] = question

    def __hash__(self) -> int:
        return hash(self.NAME)

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, Category):
            return self.NAME == other.NAME
        return False

    def getCategoryHeader(self, subCategory: str | None = None) -> ET.Element:
        """Insert an <question type='category'> before all Questions of this Category."""
        header = ET.Element("question", type="category")
        cat = ET.SubElement(header, "category")
        info = ET.SubElement(header, "info", format="html")
        catStr = (
            f"$module$/top/{self.NAME}"
            if subCategory is None
            else f"$module$/top/{self.NAME}/Question-{subCategory[2:]}_Variants"
        )
        ET.SubElement(cat, "text").text = catStr
        ET.SubElement(info, "text").text = str(self.desc)
        ET.SubElement(header, "idnumber").text = (
            f"cat-{self.id}" if subCategory is None else f"variants-{subCategory}"
        )
        ET.indent(header)
        return header



"""Numerical question implementation."""

from types import UnionType
from typing import ClassVar

import lxml.etree as ET

from excel2moodle.core.globals import (
    Tags,
    TextElements,
    XMLTags,
)
from excel2moodle.core.parser import QuestionParser
from excel2moodle.core.question import Question


class NFQuestion(Question):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    optionalTags: ClassVar[dict[Tags, type | UnionType]] = {
        Tags.BPOINTS: str,
    }
    mandatoryTags: ClassVar[dict[Tags, type | UnionType]] = {
        Tags.RESULT: float | int,
    }


class NFQuestionParser(QuestionParser):
    """Subclass for parsing numeric questions."""

    def __init__(self) -> None:
        super().__init__()
        self.feedBackList = {XMLTags.GENFEEDB: Tags.GENERALFB}

    def setup(self, question: NFQuestion) -> None:
        self.question: NFQuestion = question
        super().setup(question)

    def _parseAnswers(self) -> list[ET.Element]:
        result: float = self.rawData.get(Tags.RESULT)
        ansEle: list[ET.Element] = []
        ansEle.append(self.getNumericAnsElement(result=result))
        ansEle.append(
            self.getNumericAnsElement(
                result=-result,
                fraction=self.rawData.get(Tags.WRONGSIGNPERCENT),
                feedback=self.rawData.get(Tags.WRONGSIGNFB),
                feedbackStyle=TextElements.SPANORANGE,
            )
        )
        return ansEle

    def _finalizeParsing(self) -> None:
        return super()._finalizeParsing()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'UI_mainWindow.ui'
##
## Created by: Qt User Interface Compiler version 6.10.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QAbstractSpinBox, QApplication,
    QCheckBox, QDoubleSpinBox, QFormLayout, QFrame,
    QGridLayout, QHeaderView, QLabel, QLayout,
    QLineEdit, QMainWindow, QMenu, QMenuBar,
    QPushButton, QSizePolicy, QSpinBox, QSplitter,
    QStatusBar, QTableWidget, QTableWidgetItem, QTextEdit,
    QToolBar, QTreeWidget, QTreeWidgetItem, QVBoxLayout,
    QWidget)

class Ui_MoodleTestGenerator(object):
    def setupUi(self, MoodleTestGenerator):
        if not MoodleTestGenerator.objectName():
            MoodleTestGenerator.setObjectName(u"MoodleTestGenerator")
        MoodleTestGenerator.resize(1101, 1010)
        icon = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.DialogQuestion))
        MoodleTestGenerator.setWindowIcon(icon)
        self.actionSpreadsheet = QAction(MoodleTestGenerator)
        self.actionSpreadsheet.setObjectName(u"actionSpreadsheet")
        icon1 = QIcon(QIcon.fromTheme(u"document-open"))
        self.actionSpreadsheet.setIcon(icon1)
        self.actionEquationChecker = QAction(MoodleTestGenerator)
        self.actionEquationChecker.setObjectName(u"actionEquationChecker")
        icon2 = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.ToolsCheckSpelling))
        self.actionEquationChecker.setIcon(icon2)
        self.actionParseAll = QAction(MoodleTestGenerator)
        self.actionParseAll.setObjectName(u"actionParseAll")
        icon3 = QIcon(QIcon.fromTheme(u"view-refresh"))
        self.actionParseAll.setIcon(icon3)
        self.actionExport = QAction(MoodleTestGenerator)
        self.actionExport.setObjectName(u"actionExport")
        icon4 = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.DocumentSave))
        self.actionExport.setIcon(icon4)
        self.actionAbout = QAction(MoodleTestGenerator)
        self.actionAbout.setObjectName(u"actionAbout")
        icon5 = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.HelpAbout))
        self.actionAbout.setIcon(icon5)
        self.actionDocumentation = QAction(MoodleTestGenerator)
        self.actionDocumentation.setObjectName(u"actionDocumentation")
        self.actionDocumentation.setIcon(icon5)
        self.actionGenerateVariables = QAction(MoodleTestGenerator)
        self.actionGenerateVariables.setObjectName(u"actionGenerateVariables")
        icon6 = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.ListAdd))
        self.actionGenerateVariables.setIcon(icon6)
        self.actionCopyVariables = QAction(MoodleTestGenerator)
        self.actionCopyVariables.setObjectName(u"actionCopyVariables")
        icon7 = QIcon(QIcon.fromTheme(u"edit-copy"))
        self.actionCopyVariables.setIcon(icon7)
        self.actionOpenSpreadsheetExternal = QAction(MoodleTestGenerator)
        self.actionOpenSpreadsheetExternal.setObjectName(u"actionOpenSpreadsheetExternal")
        icon8 = QIcon(QIcon.fromTheme(u"mail-forward"))
        self.actionOpenSpreadsheetExternal.setIcon(icon8)
        self.actionLaTex = QAction(MoodleTestGenerator)
        self.actionLaTex.setObjectName(u"actionLaTex")
        self.actionLaTex.setCheckable(True)
        icon9 = QIcon(QIcon.fromTheme(QIcon.ThemeIcon.DocumentPrintPreview))
        self.actionLaTex.setIcon(icon9)
        self.mainWidget = QWidget(MoodleTestGenerator)
        self.mainWidget.setObjectName(u"mainWidget")
        self.verticalLayout_3 = QVBoxLayout(self.mainWidget)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.splitter_2 = QSplitter(self.mainWidget)
        self.splitter_2.setObjectName(u"splitter_2")
        self.splitter_2.setOrientation(Qt.Orientation.Vertical)
        self.splitter_2.setHandleWidth(3)
        self.splitter = QSplitter(self.splitter_2)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.splitter.setHandleWidth(3)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout_4 = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(0, 0, 4, 4)
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setVerticalSpacing(6)
        self.buttonExport = QPushButton(self.layoutWidget)
        self.buttonExport.setObjectName(u"buttonExport")

        self.gridLayout.addWidget(self.buttonExport, 0, 2, 1, 1)

        self.checkBoxQuestionListSelectAll = QCheckBox(self.layoutWidget)
        self.checkBoxQuestionListSelectAll.setObjectName(u"checkBoxQuestionListSelectAll")

        self.gridLayout.addWidget(self.checkBoxQuestionListSelectAll, 1, 0, 1, 1)

        self.buttonSpreadsheet = QPushButton(self.layoutWidget)
        self.buttonSpreadsheet.setObjectName(u"buttonSpreadsheet")

        self.gridLayout.addWidget(self.buttonSpreadsheet, 0, 0, 1, 1)

        self.pointCounter = QDoubleSpinBox(self.layoutWidget)
        self.pointCounter.setObjectName(u"pointCounter")
        self.pointCounter.setEnabled(True)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pointCounter.sizePolicy().hasHeightForWidth())
        self.pointCounter.setSizePolicy(sizePolicy)
        self.pointCounter.setMaximumSize(QSize(120, 16777215))
        self.pointCounter.setBaseSize(QSize(190, 0))
        font = QFont()
        font.setPointSize(12)
        font.setBold(False)
        self.pointCounter.setFont(font)
        self.pointCounter.setLayoutDirection(Qt.LayoutDirection.LeftToRight)
        self.pointCounter.setAutoFillBackground(False)
        self.pointCounter.setInputMethodHints(Qt.InputMethodHint.ImhNone)
        self.pointCounter.setWrapping(False)
        self.pointCounter.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.pointCounter.setReadOnly(True)
        self.pointCounter.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.pointCounter.setDecimals(1)
        self.pointCounter.setMaximum(9999.899999999999636)

        self.gridLayout.addWidget(self.pointCounter, 1, 2, 1, 1)

        self.line_4 = QFrame(self.layoutWidget)
        self.line_4.setObjectName(u"line_4")
        self.line_4.setFrameShape(QFrame.Shape.VLine)
        self.line_4.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line_4, 1, 3, 1, 1)

        self.questionCounter = QSpinBox(self.layoutWidget)
        self.questionCounter.setObjectName(u"questionCounter")
        self.questionCounter.setMaximumSize(QSize(120, 16777215))
        font1 = QFont()
        font1.setPointSize(12)
        self.questionCounter.setFont(font1)
        self.questionCounter.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)

        self.gridLayout.addWidget(self.questionCounter, 1, 1, 1, 1)


        self.verticalLayout_4.addLayout(self.gridLayout)

        self.treeWidget = QTreeWidget(self.layoutWidget)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setTextAlignment(2, Qt.AlignLeading|Qt.AlignVCenter);
        self.treeWidget.setHeaderItem(__qtreewidgetitem)
        self.treeWidget.setObjectName(u"treeWidget")
        self.treeWidget.setBaseSize(QSize(0, 60))
        self.treeWidget.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.treeWidget.header().setCascadingSectionResizes(True)
        self.treeWidget.header().setMinimumSectionSize(8)

        self.verticalLayout_4.addWidget(self.treeWidget)

        self.splitter.addWidget(self.layoutWidget)
        self.layoutWidget1 = QWidget(self.splitter)
        self.layoutWidget1.setObjectName(u"layoutWidget1")
        self.verticalLayout_2 = QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.verticalLayout_2.setContentsMargins(4, 0, 0, 4)
        self.label = QLabel(self.layoutWidget1)
        self.label.setObjectName(u"label")
        font2 = QFont()
        font2.setPointSize(13)
        font2.setBold(False)
        font2.setItalic(True)
        self.label.setFont(font2)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.verticalLayout_2.addWidget(self.label)

        self.formLayout_3 = QFormLayout()
        self.formLayout_3.setObjectName(u"formLayout_3")
        self.formLayout_3.setLabelAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.formLayout_3.setHorizontalSpacing(20)
        self.formLayout_3.setVerticalSpacing(5)
        self.formLayout_3.setContentsMargins(10, 6, 10, -1)
        self.questionNameLabel = QLabel(self.layoutWidget1)
        self.questionNameLabel.setObjectName(u"questionNameLabel")

        self.formLayout_3.setWidget(0, QFormLayout.ItemRole.LabelRole, self.questionNameLabel)

        self.qNameLine = QLineEdit(self.layoutWidget1)
        self.qNameLine.setObjectName(u"qNameLine")
        self.qNameLine.setReadOnly(True)

        self.formLayout_3.setWidget(0, QFormLayout.ItemRole.FieldRole, self.qNameLine)


        self.verticalLayout_2.addLayout(self.formLayout_3)

        self.line_6 = QFrame(self.layoutWidget1)
        self.line_6.setObjectName(u"line_6")
        self.line_6.setFrameShape(QFrame.Shape.HLine)
        self.line_6.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout_2.addWidget(self.line_6)

        self.previewTextEdit = QTextEdit(self.layoutWidget1)
        self.previewTextEdit.setObjectName(u"previewTextEdit")

        self.verticalLayout_2.addWidget(self.previewTextEdit)

        self.tableVariables = QTableWidget(self.layoutWidget1)
        if (self.tableVariables.columnCount() < 1):
            self.tableVariables.setColumnCount(1)
        __qtablewidgetitem = QTableWidgetItem()
        self.tableVariables.setHorizontalHeaderItem(0, __qtablewidgetitem)
        self.tableVariables.setObjectName(u"tableVariables")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.tableVariables.sizePolicy().hasHeightForWidth())
        self.tableVariables.setSizePolicy(sizePolicy1)
        self.tableVariables.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)

        self.verticalLayout_2.addWidget(self.tableVariables)

        self.splitter.addWidget(self.layoutWidget1)
        self.splitter_2.addWidget(self.splitter)
        self.layoutWidget2 = QWidget(self.splitter_2)
        self.layoutWidget2.setObjectName(u"layoutWidget2")
        self.verticalLayout = QVBoxLayout(self.layoutWidget2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setSizeConstraint(QLayout.SizeConstraint.SetMinimumSize)
        self.verticalLayout.setContentsMargins(0, 3, 0, 0)
        self.label_4 = QLabel(self.layoutWidget2)
        self.label_4.setObjectName(u"label_4")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy2)
        self.label_4.setMinimumSize(QSize(0, 20))
        self.label_4.setMaximumSize(QSize(16777215, 20))
        self.label_4.setBaseSize(QSize(0, 20))
        self.label_4.setIndent(8)

        self.verticalLayout.addWidget(self.label_4)

        self.loggerWindow = QTextEdit(self.layoutWidget2)
        self.loggerWindow.setObjectName(u"loggerWindow")
        self.loggerWindow.setEnabled(True)
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.loggerWindow.sizePolicy().hasHeightForWidth())
        self.loggerWindow.setSizePolicy(sizePolicy3)
        self.loggerWindow.setMinimumSize(QSize(0, 0))
        self.loggerWindow.setMaximumSize(QSize(16777215, 800))
        self.loggerWindow.setBaseSize(QSize(0, 29))
        self.loggerWindow.setUndoRedoEnabled(False)
        self.loggerWindow.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)

        self.verticalLayout.addWidget(self.loggerWindow)

        self.splitter_2.addWidget(self.layoutWidget2)

        self.verticalLayout_3.addWidget(self.splitter_2)

        MoodleTestGenerator.setCentralWidget(self.mainWidget)
        self.menubar = QMenuBar(MoodleTestGenerator)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1101, 26))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuTools = QMenu(self.menubar)
        self.menuTools.setObjectName(u"menuTools")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        MoodleTestGenerator.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MoodleTestGenerator)
        self.statusbar.setObjectName(u"statusbar")
        MoodleTestGenerator.setStatusBar(self.statusbar)
        self.toolBar_3 = QToolBar(MoodleTestGenerator)
        self.toolBar_3.setObjectName(u"toolBar_3")
        MoodleTestGenerator.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar_3)
        self.toolBar = QToolBar(MoodleTestGenerator)
        self.toolBar.setObjectName(u"toolBar")
        MoodleTestGenerator.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionExport)
        self.menuFile.addAction(self.actionSpreadsheet)
        self.menuFile.addAction(self.actionOpenSpreadsheetExternal)
        self.menuTools.addAction(self.actionParseAll)
        self.menuTools.addAction(self.actionEquationChecker)
        self.menuTools.addAction(self.actionGenerateVariables)
        self.menuTools.addAction(self.actionCopyVariables)
        self.menuTools.addAction(self.actionLaTex)
        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionDocumentation)
        self.toolBar_3.addAction(self.actionEquationChecker)
        self.toolBar_3.addAction(self.actionDocumentation)
        self.toolBar_3.addAction(self.actionAbout)
        self.toolBar.addAction(self.actionSpreadsheet)
        self.toolBar.addAction(self.actionParseAll)
        self.toolBar.addAction(self.actionExport)

        self.retranslateUi(MoodleTestGenerator)

        QMetaObject.connectSlotsByName(MoodleTestGenerator)
    # setupUi

    def retranslateUi(self, MoodleTestGenerator):
        MoodleTestGenerator.setWindowTitle(QCoreApplication.translate("MoodleTestGenerator", u"excel 2 moodle", None))
        self.actionSpreadsheet.setText(QCoreApplication.translate("MoodleTestGenerator", u"&Import spreadsheet", None))
#if QT_CONFIG(tooltip)
        self.actionSpreadsheet.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Open the question spreadsheet", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionSpreadsheet.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"Ctrl+O", None))
#endif // QT_CONFIG(shortcut)
        self.actionEquationChecker.setText(QCoreApplication.translate("MoodleTestGenerator", u"&Equation Checker", None))
        self.actionParseAll.setText(QCoreApplication.translate("MoodleTestGenerator", u"&Parse all Questions", None))
#if QT_CONFIG(tooltip)
        self.actionParseAll.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Parse all questions inside the spreadsheet", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionParseAll.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"Ctrl+R", None))
#endif // QT_CONFIG(shortcut)
        self.actionExport.setText(QCoreApplication.translate("MoodleTestGenerator", u"Export the question selection ", None))
#if QT_CONFIG(shortcut)
        self.actionExport.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"Ctrl+E", None))
#endif // QT_CONFIG(shortcut)
        self.actionAbout.setText(QCoreApplication.translate("MoodleTestGenerator", u"About", None))
#if QT_CONFIG(tooltip)
        self.actionAbout.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"About excel2moodle", None))
#endif // QT_CONFIG(tooltip)
        self.actionDocumentation.setText(QCoreApplication.translate("MoodleTestGenerator", u"Documentation", None))
#if QT_CONFIG(tooltip)
        self.actionDocumentation.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Open the documentation for excel2moodle", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionDocumentation.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"F1", None))
#endif // QT_CONFIG(shortcut)
        self.actionGenerateVariables.setText(QCoreApplication.translate("MoodleTestGenerator", u"Variable Generator", None))
#if QT_CONFIG(tooltip)
        self.actionGenerateVariables.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Generate new variables for the question.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionGenerateVariables.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"Ctrl+G", None))
#endif // QT_CONFIG(shortcut)
        self.actionCopyVariables.setText(QCoreApplication.translate("MoodleTestGenerator", u"Copy Variables", None))
#if QT_CONFIG(tooltip)
        self.actionCopyVariables.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Copy the variables of the question to the clipboard", None))
#endif // QT_CONFIG(tooltip)
        self.actionOpenSpreadsheetExternal.setText(QCoreApplication.translate("MoodleTestGenerator", u"Open spreadsheet in external app", None))
#if QT_CONFIG(tooltip)
        self.actionOpenSpreadsheetExternal.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Open the imported spreadsheet in external application such as LibreOffice Calc", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionOpenSpreadsheetExternal.setShortcut(QCoreApplication.translate("MoodleTestGenerator", u"Ctrl+Shift+O", None))
#endif // QT_CONFIG(shortcut)
        self.actionLaTex.setText(QCoreApplication.translate("MoodleTestGenerator", u"Render LaTex", None))
#if QT_CONFIG(tooltip)
        self.actionLaTex.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Render LaTex code in the question preview. This makes the preview a bit less responsive.", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.buttonExport.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"Export the selected questions  to a xml file", None))
#endif // QT_CONFIG(tooltip)
        self.buttonExport.setText(QCoreApplication.translate("MoodleTestGenerator", u"Export", None))
        self.checkBoxQuestionListSelectAll.setText(QCoreApplication.translate("MoodleTestGenerator", u"Select all", None))
        self.buttonSpreadsheet.setText(QCoreApplication.translate("MoodleTestGenerator", u"Open Spreadsheet", None))
#if QT_CONFIG(tooltip)
        self.pointCounter.setToolTip(QCoreApplication.translate("MoodleTestGenerator", u"The total number of points of all selected questions.", None))
#endif // QT_CONFIG(tooltip)
        self.pointCounter.setPrefix("")
        self.pointCounter.setSuffix(QCoreApplication.translate("MoodleTestGenerator", u"  Pts.", None))
        self.questionCounter.setSuffix(QCoreApplication.translate("MoodleTestGenerator", u"   Qst.", None))
        ___qtreewidgetitem = self.treeWidget.headerItem()
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("MoodleTestGenerator", u"Variants", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("MoodleTestGenerator", u"Points", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("MoodleTestGenerator", u"Description", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("MoodleTestGenerator", u"Question ID", None));
        self.label.setText(QCoreApplication.translate("MoodleTestGenerator", u"Question Preview", None))
        self.questionNameLabel.setText(QCoreApplication.translate("MoodleTestGenerator", u"Question:", None))
        ___qtablewidgetitem = self.tableVariables.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("MoodleTestGenerator", u"Variable", None));
        self.label_4.setText(QCoreApplication.translate("MoodleTestGenerator", u"Logging Messages", None))
        self.menuFile.setTitle(QCoreApplication.translate("MoodleTestGenerator", u"File", None))
        self.menuTools.setTitle(QCoreApplication.translate("MoodleTestGenerator", u"Tools", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MoodleTestGenerator", u"Help", None))
        self.toolBar_3.setWindowTitle(QCoreApplication.translate("MoodleTestGenerator", u"toolBar_3", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("MoodleTestGenerator", u"toolBar", None))
    # retranslateUi




from pathlib import Path

import pytest

from excel2moodle.core.dataStructure import QuestionDB
from excel2moodle.core.settings import Settings, Tags

settings = Settings()
excelFile = Path("test/TestQuestion.ods").resolve()
katName = "NFM2"
database = QuestionDB(settings)

settings.set(Tags.QUESTIONVARIANT, 1)
database.spreadsheet = excelFile
database.readCategoriesMetadata(excelFile)
category = database.initCategory(katName, sheetPath=excelFile)
question = database.setupAndParseQuestion(category=category, qNumber=1)
questionData = question.rawData

CATTOLERANCE = 0.05


@pytest.mark.parametrize(
    ("tag", "expectedType"),
    [
        (Tags.BPOINTS, list),
        (Tags.TEXT, list),
        (Tags.ANSPICWIDTH, int),
        (Tags.POINTS, float),
        (Tags.MEDIASCRIPTS, list),
    ],
)
def test_validatorQuestionDataGeneration(tag, expectedType) -> None:
    assert type(questionData.get(tag)) == expectedType


def test_readOneLenghtListFromSettings() -> None:
    assert len(settings.values[Tags.MEDIASCRIPTS]) == 1
    assert type(settings.values[Tags.MEDIASCRIPTS]) is list


@pytest.mark.parametrize(
    ("tag", "expectedLiteral"),
    [
        (Tags.TYPE, ("MC", "NFM", "CLOZE")),
    ],
)
def test_literalReturs(tag, expectedLiteral) -> None:
    assert questionData.get(tag) in expectedLiteral


def test_ReturnTypeWhileIterating() -> None:
    for tag in Tags:
        if tag.place == "project":
            val = questionData.get(tag)
            if val:
                assert type(val) == tag.typ()


@pytest.mark.parametrize(
    ("tag", "value"),
    [
        (Tags.POINTS, 5.0),
        (Tags.PICTUREWIDTH, 650),
        (Tags.WRONGSIGNPERCENT, 30),
        (Tags.TOLERANCE, CATTOLERANCE),
    ],
)
def test_usageOfCategorySettings(tag, value) -> None:
    assert questionData.get(tag) == value


@pytest.mark.parametrize(
    ("val", "expectation"),
    [
        (-1, CATTOLERANCE),
        (0, CATTOLERANCE),
        (23, 0.23),
        (100, CATTOLERANCE),
    ],
)
def test_manipulateInvaldTolerance(val, expectation) -> None:
    questionData.categoryFallbacks[Tags.TOLERANCE] = val
    assert questionData.get(Tags.TOLERANCE) == expectation



from pathlib import Path

import pytest

from excel2moodle.core.question import Picture

imgFolder = Path("./test/Abbildungen")

qID = "0103"
katName = "MC1"


@pytest.mark.parametrize(
    ("imgKey", "path", "width", "height"),
    [
        ("1:height:100", "0101.png", 0, 100),
        ("true", "0103.svg", 300, 0),
        ("yes", "0103.svg", 300, 0),
        ("1_a:width:100", "0101_a.png", 100, 0),
        ("0101:height:255", "0101.png", 0, 255),
        ("03:width:545", "0103.svg", 545, 0),
        ("101_a:height:80", "0101_a.png", 0, 80),
        pytest.param(
            "101_a:height:80:width:100", "0101_a.png", 100, 80, marks=pytest.mark.xfail
        ),
    ],
)
def test_PictureFindImgFile(imgKey, path, width, height) -> None:
    imgF = (imgFolder / katName).resolve()
    picture = Picture(imgKey, imgF, qID, width=300)
    print(picture.path)
    p = str(picture.path.stem + picture.path.suffix)
    size = picture.size
    assert p == path
    assert size.get("height", "0") == str(height)
    assert size.get("width", "0") == str(width)


@pytest.mark.parametrize(
    ("imgKey", "expected"),
    [
        ("2_b", "0102_b"),
        ("01_a", "0101_a"),
        ("0101", "0101"),
        ("05-c", "0105-c"),
        ("201", "0201"),
        ("101_a", "0101_a"),
        ("3802_c", "3802_c"),
        ("0902_c", "0902_c"),
        pytest.param("13802_c", "3802_c", marks=pytest.mark.xfail),
        pytest.param("false", None, marks=pytest.mark.xfail),
    ],
)
def test_Picture_EvaluateCorrectPicID(imgKey, expected) -> None:
    imgF = (imgFolder / katName).resolve(strict=True)
    picture = Picture(imgKey, imgF, qID, width=300)
    assert picture.picID == expected



"""Main Module which does the heavy lifting.

At the heart is the class ``xmlTest``
"""

import datetime as dt
import logging
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import TYPE_CHECKING

import lxml.etree as ET  # noqa: N812
import pandas as pd
import yaml
from PySide6.QtCore import QObject, Signal
from PySide6.QtWidgets import QDialog

from excel2moodle import __version__
from excel2moodle.core import stringHelpers
from excel2moodle.core.category import Category
from excel2moodle.core.exceptions import InvalidFieldException, QNotParsedException
from excel2moodle.core.globals import Tags
from excel2moodle.core.question import ParametricQuestion, Question
from excel2moodle.core.settings import Settings
from excel2moodle.core.validator import Validator
from excel2moodle.logger import LogAdapterQuestionID
from excel2moodle.question_types import QuestionTypeMapping
from excel2moodle.question_types.cloze import ClozeQuestion, ClozeQuestionParser
from excel2moodle.question_types.mc import MCQuestion, MCQuestionParser
from excel2moodle.question_types.nf import NFQuestion, NFQuestionParser
from excel2moodle.question_types.nfm import NFMQuestion, NFMQuestionParser
from excel2moodle.ui.dialogs import QuestionVariantDialog
from excel2moodle.ui.treewidget import QuestionItem

if TYPE_CHECKING:
    from PySide6.QtWidgets import QMainWindow

logger = logging.getLogger(__name__)


class QuestionDBSignals(QObject):
    categoryReady = Signal(Category)
    categoryQuestionsReady = Signal(Category)


def processSheet(sheetPath: str, categoryName: str) -> pd.DataFrame:
    """Parse `categoryName` from the file ``sheetPath`` into the dataframe.

    This Function is meant to be run asynchron for increased speed.
    """
    return pd.read_excel(
        Path(sheetPath),
        sheet_name=str(categoryName),
        index_col=0,
        header=None,
        engine="calamine",
    )


class QuestionDB:
    """The QuestionDB is the main class for processing the Spreadsheet.

    It provides the functionality, for setting up the categories and Questions.
    Any interaction with the questions are done by its methods.
    """

    signals = QuestionDBSignals()
    validator: Validator = Validator()
    nfParser: NFQuestionParser = NFQuestionParser()
    nfmParser: NFMQuestionParser = NFMQuestionParser()
    mcParser: MCQuestionParser = MCQuestionParser()
    clozeParser: ClozeQuestionParser = ClozeQuestionParser()

    def __init__(self, settings: Settings) -> None:
        self.settings = settings
        self.window: QMainWindow | None = None
        self.categoriesMetaData: pd.DataFrame
        self.categories: dict[str, Category]
        self._exportedQuestions: list[Question] = []
        self._exportedAll: bool = False

    @property
    def spreadsheet(self) -> Path:
        return self._spreadsheet

    @spreadsheet.setter
    def spreadsheet(self, sheet) -> None:
        self.settings.clear()
        self._spreadsheet = sheet
        self.settings.set(Tags.SPREADSHEETPATH, sheet)
        logger.info("saved new spreadsheet %s", sheet)

    def readCategoriesMetadata(self, sheetPath: Path | None = None) -> pd.DataFrame:
        """Read the metadata and questions from the spreadsheet.

        Get the category data from the spreadsheet and stores it in the
        ``categoriesMetaData`` dataframe
        Setup the categories and store them  in ``self.categories = {}``
        Pass the question data to the categories.

        Raises
        ------
        ValueError
            When there is no 'seetings' worksheet in the file.
        InvalidFieldException
            When the settings are invalid
            Or When the categories Sheet doesn't provide the necessary keys.

        Before raising it logges the exceptions with a meaningful message.

        """
        sheetPath = sheetPath if sheetPath else self.spreadsheet
        logger.info("Start Parsing the Excel Metadata Sheet\n")
        with Path(sheetPath).open("rb") as f:
            settingDf = pd.read_excel(
                f,
                sheet_name="settings",
                index_col=0,
                header=None,
                engine="calamine",
            )
        logger.debug("Found the settings: \n\t%s", settingDf)
        settingDf = self.harmonizeDFIndex(settingDf)
        settingsDict = Validator.dfToDict(settingDf[1])
        Validator.listify(settingsDict)
        for tag, value in settingsDict.items():
            self.settings.set(tag, value)

        self._validateProjectSettings(sheetPath=sheetPath)
        with Path(sheetPath).open("rb") as f:
            self.categoriesMetaData = pd.read_excel(
                f,
                sheet_name=self.settings.get(Tags.CATEGORIESSHEET),
                index_col=0,
                engine="calamine",
            )
        if "description" not in self.categoriesMetaData.columns:
            msg = f"You need to specify the 'description' tag for each category in the sheet '{self.settings.get(Tags.CATEGORIESSHEET)}'."
            raise InvalidFieldException(msg, "0000", "description")
        logger.info("Sucessfully read categoriesMetaData")
        return self.categoriesMetaData

    def _validateProjectSettings(self, sheetPath: Path) -> None:
        if Tags.LOGLEVEL in self.settings:
            level: str = self.settings.get(Tags.LOGLEVEL)
            if level.upper() not in ("DEBUG", "INFO", "WARNING", "ERROR"):
                self.settings.pop(Tags.LOGLEVEL)
                logger.warning("You specified an unsupported Loglevel: %s", level)
        if self.window is not None:
            self.window.logHandler.setLevel(self.settings.get(Tags.LOGLEVEL).upper())
        if Tags.IMPORTMODULE in self.settings:
            logger.warning(
                "Appending: %s to sys.path. All names defined by it will be usable",
                sheetPath.parent,
            )
            sys.path.append(str(sheetPath.parent))
        if Tags.PICTURESUBFOLDER not in self.settings:
            logger.warning("You didn't specify an image Folder. This may cause errors.")
        imgFolder = self.settings.get(Tags.SPREADSHEETPATH).parent / self.settings.get(
            Tags.PICTURESUBFOLDER
        )
        catSheet = self.settings.get(Tags.CATEGORIESSHEET)
        if catSheet not in pd.ExcelFile(sheetPath, engine="calamine").sheet_names:
            msg = f"The specified categories sheet: '{catSheet}' doesn't exist."
            raise InvalidFieldException(msg, "00000", Tags.CATEGORIESSHEET)
        try:
            imgFolder.resolve(strict=True)
        except FileNotFoundError:
            msg = f"Img Path: {imgFolder} could not be found"
            raise InvalidFieldException(msg, "00000", Tags.PICTURESUBFOLDER)
        else:
            self.settings.set(Tags.PICTUREFOLDER, imgFolder.resolve())
            logger.info("Set up the project settings")

    def initAllCategories(self, sheetPath: Path | None = None) -> None:
        """Read all category sheets and initialize all Categories."""
        sheetPath = sheetPath if sheetPath else self.spreadsheet
        if not hasattr(self, "categoriesMetaData"):
            logger.error("Can't process the Categories without Metadata")
            return
        if hasattr(self, "categories"):
            self.categories.clear()
        else:
            self.categories: dict[str, Category] = {}
        with pd.ExcelFile(sheetPath, engine="calamine") as excelFile:
            for categoryName in excelFile.sheet_names:
                logger.debug("Starting to read category %s", categoryName)
                if categoryName in self.categoriesMetaData.index:
                    self.initCategory(categoryName, sheetPath=sheetPath)

    def asyncInitAllCategories(self, sheetPath: Path | None = None) -> None:
        """Read all category sheets asynchron and initialize all Categories.

        It does the same as `initAllCategories` but the parsing of the excelfile
        is done asynchron via `concurrent.futures.ProcessPoolExecutor`
        """
        sheetPath = sheetPath if sheetPath else self.spreadsheet
        if not hasattr(self, "categoriesMetaData"):
            logger.error("Can't process the Categories without Metadata")
            return
        if hasattr(self, "categories"):
            self.categories.clear()
        else:
            self.categories: dict[str, Category] = {}
        sheetNames = []
        with pd.ExcelFile(sheetPath, engine="calamine") as excelFile:
            sheetNames = [
                name
                for name in excelFile.sheet_names
                if name in self.categoriesMetaData.index
            ]
        logger.debug("found those category sheets: \n %s ", sheetNames)
        with ProcessPoolExecutor() as executor:
            futures = {
                executor.submit(processSheet, str(sheetPath), sheet): sheet
                for sheet in sheetNames
            }
            for future in as_completed(futures):
                categoryName = futures[future]
                try:
                    categoryDataF = future.result()
                    self._setupCategory(categoryDataF, categoryName)
                    logger.debug("Finished processing %s", categoryName)
                except Exception as e:
                    logger.exception("Error processing sheet %s: %s", categoryName, e)
                    logger.debug("Future exception: %s", future.exception())

    def initCategory(
        self, categoryName: str, sheetPath: Path | None = None
    ) -> bool | Category:
        """Read `categoryName` from the ``sheetPath`` and initialize the category.
        Returns the Category and appends it to `self.categories`.
        """
        sheetPath = sheetPath if sheetPath else self.spreadsheet
        katDf = pd.read_excel(
            sheetPath,
            sheet_name=str(categoryName),
            index_col=0,
            header=None,
            engine="calamine",
        )
        if not katDf.empty:
            logger.debug("Sucessfully read the Dataframe for cat %s", categoryName)
            return self._setupCategory(katDf, categoryName)
        return False

    @staticmethod
    def harmonizeDFIndex(dataframe: pd.DataFrame) -> pd.DataFrame:
        """Convert the index strings to lowercase without whitespace."""
        index = dataframe.index.str.lower()
        harmonizedIdx = ["".join(i.split()) if pd.notna(i) else i for i in index]
        dataframe.index = pd.Index(harmonizedIdx)
        return dataframe

    def _setupCategory(self, categoryDf: pd.DataFrame, categoryName: str) -> Category:
        """Setup the category from the ``dataframe``.

        :emits: categoryReady(self) Signal.
        """
        categoryDf: pd.DataFrame = self.harmonizeDFIndex(categoryDf)
        categorySettings: dict[str, float | int | str] = self.harmonizeDFIndex(
            self.categoriesMetaData.loc[categoryName]
        ).to_dict()
        nanSettings: list[str] = [
            k for k, v in categorySettings.items() if pd.isna(v) or k not in Tags
        ]
        for na in nanSettings:
            categorySettings.pop(na)
        category = Category(
            categoryName,
            self.categoriesMetaData["description"].loc[categoryName],
            dataframe=categoryDf,
            settings=categorySettings,
        )
        if hasattr(self, "categories"):
            self.categories[categoryName] = category
            self.signals.categoryReady.emit(category)
        else:
            logger.warning("Can't append initialized category to the database")
        logger.debug("Category %s is initialized", categoryName)
        return category

    def parseAllQuestions(self) -> None:
        """Parse all question from all categories.

        The categories need to be initialized first.
        """
        for category in self.categories.values():
            self.parseCategoryQuestions(category)

    def parseCategoryQuestions(self, category: Category) -> None:
        """Parse all questions inside ``category``.

        The category has to be initialized first.
        """
        for qNum in category.dataframe.columns:
            try:
                self.setupAndParseQuestion(category, qNum)
            except (InvalidFieldException, QNotParsedException, AttributeError):
                logger.exception(
                    "Question %s%02d couldn't be parsed. The Question Data: \n %s",
                    category.id,
                    qNum,
                    category.dataframe[qNum],
                )
        self.signals.categoryQuestionsReady.emit(category)

    @classmethod
    def setupAndParseQuestion(cls, category: Category, qNumber: int) -> Question:
        """Check if the Question Data is valid. Then parse it.

        The Question data is accessed from  `category.dataframe` via its number
        First it is checked if all mandatory fields for the given question type
        are provided.
        Then in checks, weather the data has the correct type.
        If the data is valid, the corresponding parser is fed with the data and run.

        Raises
        ------
        QNotParsedException
            If the parsing of the question is not possible this is raised
        InvalidFieldException
            If the data of the question is invalid.
            This gives more information wheather a missing field, or the invalid type
            caused the Exception.

        """
        locallogger = LogAdapterQuestionID(
            logger, {"qID": f"{category.id}{qNumber:02d}"}
        )
        locallogger.debug("Starting to check Validity")
        qdat = category.dataframe[qNumber]
        if not isinstance(qdat, pd.Series):
            locallogger.error("cannot validate data that isn't a pd.Series")
            msg = "cannot validate data that isn't a pd.Series"
            raise QNotParsedException(msg, f"{category.id}{qNumber}")
        cls.validator.setup(qdat, qNumber)
        cls.validator.validate()
        validData = cls.validator.getQuestionData()
        qtype: str = validData.get(Tags.TYPE)
        logger.debug("Question type is: %s", qtype)
        question = QuestionTypeMapping[qtype].create(category, validData)
        if question.isParsed:
            locallogger.info("Question already parsed")
            return question
        if isinstance(question, NFQuestion):
            cls.nfParser.setup(question)
            locallogger.debug("setup a new NF parser ")
            cls.nfParser.parse()
        elif isinstance(question, MCQuestion):
            cls.mcParser.setup(question)
            locallogger.debug("setup a new MC parser ")
            cls.mcParser.parse()
        elif isinstance(question, NFMQuestion):
            cls.nfmParser.setup(question)
            locallogger.debug("setup a new NFM parser ")
            cls.nfmParser.parse()
        elif isinstance(question, ClozeQuestion):
            cls.clozeParser.setup(question)
            locallogger.debug("setup a new CLOZE parser")
            cls.clozeParser.parse()
        else:
            msg = "couldn't setup Parser"
            raise QNotParsedException(msg, question.id)
        category.appendQuestion(qNumber, question)
        return question

    def appendQuestions(
        self,
        questions: list[QuestionItem],
        file: Path | None = None,
        pCount: int = 0,
        qCount: int = 0,
    ) -> None:
        """Append selected question Elements to the tree."""
        self._exportedQuestions.clear()
        tree = ET.Element("quiz")
        catdict: dict[Category, list[Question]] = {}
        for q in questions:
            logger.debug(f"got a question to append {q=}")
            cat = q.parent().category
            if cat not in catdict:
                catdict[cat] = []
            catdict[cat].append(q.question)
        for cat, qlist in catdict.items():
            self._appendQElements(
                cat,
                qlist,
                tree=tree,
                includeHeader=self.settings.get(Tags.INCLUDEINCATS),
            )
        stringHelpers.printDom(tree, file=file)
        if self.settings.get(Tags.GENEXPORTREPORT):
            self.generateExportReport(file, pCount=pCount, qCount=qCount)

    def _appendQElements(
        self,
        cat: Category,
        qList: list[Question],
        tree: ET.Element,
        includeHeader: bool = True,
    ) -> None:
        variant: int = self.settings.get(Tags.QUESTIONVARIANT)
        if includeHeader or variant == -1:
            tree.append(cat.getCategoryHeader())
            logger.debug(f"Appended a new category item {cat=}")
            self._exportedAll: bool = True
        for q in qList:
            if not isinstance(q, ParametricQuestion):
                tree.append(q.getUpdatedElement())
                self._exportedQuestions.append(q)
                continue
            if variant == -1:
                tree.append(cat.getCategoryHeader(subCategory=q.id))
                for var in range(q.parametrics.variants):
                    tree.append(q.getUpdatedElement(variant=var))
            elif variant == 0 or variant > q.parametrics.variants:
                dialog = QuestionVariantDialog(self.window, q)
                if dialog.exec() == QDialog.Accepted:
                    variant = dialog.variant
                    logger.debug("Die Fragen-Variante %s wurde gewählt", variant)
                else:
                    logger.warning("Keine Fragenvariante wurde gewählt.")
                tree.append(q.getUpdatedElement(variant=variant))
            else:
                tree.append(q.getUpdatedElement(variant=variant))
            self._exportedQuestions.append(q)

    def generateExportReport(
        self, file: Path | None = None, pCount: int = 0, qCount: int = 0
    ) -> None:
        """Generate a YAML report of the exported questions."""
        if not self._exportedQuestions:
            return
        if file:
            base_path = file.with_name(f"{file.stem}_export_report.yaml")
        else:
            base_path = self.spreadsheet.parent / "export_report.yaml"

        for i in range(99):
            report_path = base_path.with_name(f"{base_path.stem}-{i:02d}.yaml")
            if not report_path.resolve().exists():
                break

        report_data = {
            "export_metadata": {
                "export_time": dt.datetime.now(tz=None).strftime("%Y-%m-%d %H:%M:%S"),
                "excel2moodle_version": __version__,
            },
            "categories": {},
        }
        if qCount != 0:
            report_data["export_metadata"]["question_count"] = qCount
        if pCount != 0:
            report_data["export_metadata"]["total_point_count"] = pCount

        sorted_questions = sorted(
            self._exportedQuestions, key=lambda q: (q.category.name, q.id)
        )

        for question in sorted_questions:
            category_name = question.category.name
            if category_name not in report_data["categories"]:
                report_data["categories"][category_name] = {
                    "description": question.category.desc,
                    "questions": [],
                }

            question_data = {"id": question.id, "name": question.name}
            if isinstance(question, ParametricQuestion) and question.currentVariant > 0:
                if self._exportedAll:
                    question_data["exported_variant"] = "all"
                else:
                    question_data["exported_variant"] = question.currentVariant + 1

            report_data["categories"][category_name]["questions"].append(question_data)

        with report_path.open("w") as f:
            yaml.dump(report_data, f, sort_keys=False)
